package nxm.sys.libg;

import java.awt.Color;

import nxm.sys.inc.Constants;
import nxm.sys.inc.InternalUseOnly;
import nxm.sys.inc.MidasColor;
import nxm.sys.lib.Convert;
import nxm.sys.lib.Data;
import nxm.sys.lib.Position;
import nxm.sys.lib.Table;
import nxm.sys.lib.Time;

/** The Point3D class holds the functionality to create a Point/Ellipse/Cone
 *  This class was extracted out of LayerT5. It is default access for use
 *  only internally.
 *
 *  @see Region
 *  @see RegionGroup
 *  @see LayerT5
 *  @author Jeanette McCall / Kevin Esterheld
 *  @since NeXtMidas 3.9.3 - previously in LayerT5 as private internal class.
 */
class Point3D {

  String name = ""; // The name of the point

  private String note = ""; // The notes for the point

  int indexInRegion;
  int indexInPointsObject;
  boolean posSet = false; // Has POS been set?
  int distFromCurrentPix = -1; // Used when selecting a point. It is available in case there are multiple LayerT5s that need to be checked for closest point since 3.9.3
  double posAlt = 0.0; // The POS position (alt,lat,lon)

  double posLat = 0.0; // The POS position (alt,lat,lon)

  double posLon = 0.0; // The POS position (alt,lat,lon)

  boolean pospSet = false; // Has POSP been set?

  double pospAlt = 0.0; // The POSP position (alt,lat,lon)

  double pospLat = 0.0; // The POSP position (alt,lat,lon)

  double pospLon = 0.0; // The POSP position (alt,lat,lon)

  boolean elpsSet = false; // Has ellipse been set?

  double elpsSemiMajor = 0.0; // The ellipse parameters (major,minor,tilt)

  double elpsSemiMinor = 0.0; // The ellipse parameters (major,minor,tilt)

  double elpsTilt = 0.0; // The ellipse parameters (major,minor,tilt)

  boolean coneSet = false; // Has cone been set?

  double coneA = 0.0; // The cone parameters (radius,angle,angle)

  double coneB = 0.0; // The cone parameters (radius,angle,angle)

  double coneAngle = 0.0; // The cone parameters (radius,angle,angle)

  boolean conn = false; // The connect flags setting the drawing mode

  float hdg = 0.0f; // Symbol Heading

  byte flagNameLen = 0; // Flag 0: Name length in characters)

  byte flagNamePos = 0; // Flag 1: Name position (not supported)

  byte flagFgColor = 0; // Flag 2: foreground color (X-Midas color index number)

  byte flagBgColor = 0; // Flag 3: background color (not supported)

  byte flagCone = 0; // Flag 4: Cone/ellipse flag (0=hidden, -9=visible)

  byte flagInvert = 0; // Flag 5: Invert (not supported)

  byte flagSymCode = 0; // Flag 6: Symbol (X-Midas symbol ID number)

  byte flagSymSize = 0; // Flag 7: Symbol size

  int _ixPos = 0; // Only used by draw() - Pixel X for pos

  int _iyPos = 0; // Only used by draw() - Pixel Y for pos

  int _ixPosp = 0; // Only used by draw() - Pixel X for posp

  int _iyPosp = 0; // Only used by draw() - Pixel Y for posp

  private double[] coords = null; // Only used by prep(..)

  double minLon = 0.0; // Internal use only: max longitude used in hit(..)

  double maxLon = 0.0; // Internal use only: max longitude used in hit(..)

  double minLat = 0.0; // Internal use only: min latitude used in hit(..)

  double maxLat = 0.0; // Internal use only: min latitude used in hit(..)

  double minAlt = 0.0; // Internal use only: min altitude used in hit(..)

  double maxAlt = 0.0; // Internal use only: min altitude used in hit(..)

  private double lonNormalization = 0.0; // amount of adjustment made when normalizing POS longitude
  private double latNormalization = 0.0; // amount of adjustment made when normalizing POS latitude

  double xRangeMax = 0.0; // Stores Max of abs(minLon) and maxLon - used when plotting points on 3D graph
  double yRangeMax = 0.0; // Stores Max of abs(minLat) and maxLon - used when plotting points on 3D graph
  double zRangeMax = 0.0; // Stores Max of abs(minAlt) and maxLon - used when plotting points on 3D graph

  private boolean hidden;

  @InternalUseOnly
  public int xIndexInRegionsArray = -1; // EXPERIMENTAL

  @InternalUseOnly
  public int yIndexInRegionsArray = -1; // EXPERIMENTAL

  public LayerT5 layer;
  double time = 0;
  Color directColor = null;
  Object metadata = null;
  private MPlot MP;

  private boolean isGeodeticData = true;
  boolean in2DViewRange = false;

  /** Creates a new instance. */
  public Point3D(LayerT5 layer) {
    coords = new double[6];
    if (layer!=null) setLayerFields(layer);
  }

  /** Creates a new instance. */
  Point3D (Table tbl, LayerT5 layer) {
    this(layer);
    this.fromTable(tbl);
    if (layer!=null) setLayerFields(layer);
  }

  private void setLayerFields (LayerT5 layer) {
    this.layer = layer;
    MP = layer.getMPlot();
    this.isGeodeticData = layer.isGeodeticData();

  }

  /** Prepare for use inside draw() -- only used by draw() method. */
  void prep (boolean doPosp) {
    int k = 0;
    if (doPosp) {
      k = MP.viewer.rwc2pix(coords, 0, 3, 2, layer.pix);
      _ixPosp = layer.pix.x[1];
      _iyPosp = layer.pix.y[1];
    }
    else {
      k = MP.viewer.rwc2pix(coords, 0, 3, 1, layer.pix);
    }
    hidden = (MP.view == MPlot.V_GEODETIC) && k < 1;
    _ixPos = layer.pix.x[0];
    _iyPos = layer.pix.y[0];
  }

  /** Done using in draw() -- only used by draw() method. */
  void done () {
    // may need to do something here in the future
  }

  /** Is this point visible -- only used by draw() method. */
  boolean hit (double lon1, double lon2, double lat1, double lat2) {
    if (coneSet)
      return true; // JC4: Not sure how to optimize this one

    if ((lat2 < minLat) || (lat1 > maxLat))
      return false; // Out of range (Lat)
    if ((lon2 < minLon) || (lon1 > maxLon)) {
      if (MP.view != MPlot.V_LATLONCONTINUOUS) return false; // Out of range (Lon)
      // Need more sophisticated range check for continuous lat lon - since 3.9.0 for NXM-3284
      if ((lon2 < minLon) && !(lon1 + MPlot.LONGITUDE_RANGE < MP.rx2)) return false;
    }
    return true;
  }

  /** Updates the min/max values for the point. */
  private void updateMinMax () {
    minLon = posLon;
    maxLon = posLon;
    minLat = posLat;
    maxLat = posLat;
    minAlt = posAlt;
    maxAlt = posAlt;
    if (pospSet) {
      minLon = Math.min(minLon, pospLon);
      maxLon = Math.max(maxLon, pospLon);
      minLat = Math.min(minLat, pospLat);
      maxLat = Math.max(maxLat, pospLat);
      minAlt = Math.min(minAlt, pospAlt);
      maxAlt = Math.max(maxAlt, pospAlt);
    }
    if (elpsSet) {
      double rad = Math.max(elpsSemiMajor, elpsSemiMinor);
      double radLat = rad * Constants.METERS2DEG * 2.0; // JC4: Add factor of 2 since this is an est.
      double radLon = rad * Constants.METERS2DEG * 2.0; // JC4: Add factor of 2 since this is an est.
      minLon = Math.min(minLon, posLon - radLon);
      maxLon = Math.max(maxLon, posLon + radLon);
      minLat = Math.min(minLat, posLat - radLat);
      maxLat = Math.max(maxLat, posLat + radLat);
    }
    // Save off the maximum x,y, & z distance from 0,0,0 - used by 3D Graph view
    xRangeMax = Math.max(maxLon, Math.abs(minLon));
    yRangeMax = Math.max(maxLat, Math.abs(minLat));
    zRangeMax = Math.max(maxAlt, Math.abs(minAlt));
    layer.updateRange(this);
  }

  /** Gets the pixel position of pos. */
  Pix getPosPix (Pix p) {
    double[] pos = new double[] { posLon, posLat, posAlt };
    MP.viewer.rwc2pix(pos, 0, 3, 1, p);
    return p;
  }

  /** Gets the pixel position of posp. */
  private Pix getPospPix (Pix p) {
    double[] posp = new double[] { pospLon, pospLat, pospAlt };
    MP.viewer.rwc2pix(posp, 0, 3, 1, p);
    return p;
  }

  /** Sets the name of the point. */
  public void setName (String name) {
    this.name = name;
  }

  /** Gets the name of the point. */
  public String getName () {
    return name;
  }

  /** Sets the name of the point. */
  public void setNote (String note) {
    this.note = note;
  }

  /** Gets the name of the point. */
  public String getNote () {
    return note;
  }

  /** Gets the connection flag that indicates if this point is connected to the previous one. */
  boolean getConn () {
    return conn;
  }

  /** Sets the connection flag to indicate if this point is connected to the previous one. */
  void setConn (boolean conn) {
    this.conn = conn;
  }

  /** Sets the connection flag to indicate if this point is connected to the previous one. */
  void setConn (byte conn) {
    this.conn = (conn == 1);
  }

  /** Gets the heading of the symbol. */
  float getHdg () {
    return hdg;
  }

  /** Sets the heading of the symbol. */
  void setHdg (float hdg) {
    this.hdg = hdg;
  }

  /** Sets the connection flag to indicate if this point is connected to the previous one. */
  private void setConn (Object obj) {
    if (obj instanceof Data)
      obj = toArray(obj, new byte[1]);
    if (obj instanceof byte[])
      setConn(((byte[]) obj)[0] != 0);
    else if (obj != null)
      setConn(Convert.s2state(obj.toString(), conn));
  }

  /** Sets the direct color specification */
  void setDirectColor (Object obj) {
    directColor = MidasColor.getColor(obj);
  }

  /** Gets the direct color setting for the point. */
  Color getColor () {
    return directColor;
  }

  /** Sets the direct color specification */
  void setColor (Object obj) {
    directColor = MidasColor.getColor(obj);
  }

  /** Gets the direct color setting for the point. */
  Color getDirectColor () {
    return directColor;
  }

  /**
   * Sets the position of the point.
   * @param alt The altitude.
   * @param lat The latitude (in deg).
   * @param lon The longitude (in deg).
   */
  public void setPos (double alt, double lat, double lon) {
    if (isGeodeticData) {
      // wrap points for NXM-3510  3.9.3
      lonNormalization = 0;
      latNormalization = 0;
      while (lon > MPlot.LONGITUDE_DEFAULT_MAX) {lon -= MPlot.LONGITUDE_RANGE; lonNormalization -= MPlot.LONGITUDE_RANGE;}
      while (lon < MPlot.LONGITUDE_DEFAULT_MIN) {lon += MPlot.LONGITUDE_RANGE; lonNormalization += MPlot.LONGITUDE_RANGE;}
      while (lat > MPlot.LATITUDE_DEFAULT_MAX)  {lat -= MPlot.LATITUDE_RANGE;  latNormalization -= MPlot.LATITUDE_RANGE;}
      while (lat < MPlot.LATITUDE_DEFAULT_MIN)  {lat += MPlot.LATITUDE_RANGE;  latNormalization += MPlot.LATITUDE_RANGE;}
    }
    coords[0] = posLon = lon;
    coords[1] = posLat = lat;
    coords[2] = posAlt = alt;
    posSet = true;
    updateMinMax();
  }

  /** Internal use only */
  private void setPos (Object obj) {
    if (obj instanceof Table)
      setPos((Table) obj);
    if (obj instanceof Position)
      setPos((Position) obj);
    if (obj instanceof double[])
      setPos((double[]) obj);
  }

  /**
   * Sets the position of the point with a table.
   * @param t Pointing position table <code>{LON=lon,LAT=lat,ALT=alt}</code>
   */
  public void setPos (Table t) {
    setPos(t.getD("ALT", posAlt), t.getD("LAT",posLat), t.getD("LON",posLon)); // default to exiting values - since 3.7.0 Bug 2965
  }

  /**
   * Sets the position of the point with a Position.
   * @param p Position in {@link Position#GEODETIC} mode.
   */
  public void setPos (Position p) {
    setPos(p.getAlt(), p.getLat(), p.getLon());
  }

  /**
   * Sets the position of the point with an array (lon,lat,alt).
   * @param a The pointing position <code>{ lon, lat, alt }</code>
   */
  public void setPos (double[] a) {
    setPos(a[2], a[1], a[0]);
  }

  /** Gets the position. */
  public Position getPos () {
    return Position.fromGeo(posAlt, posLat, posLon);
  }

  /**
   * Sets the pointing position of the point.
   * @param alt The altitude.
   * @param lat The latitude (in deg).
   * @param lon The longitude (in deg).
   */
  public void setPosp (double alt, double lat, double lon) {
    if (isGeodeticData) {
      // wrap/adjust points the same amount as for POS for NXM-3510  3.9.3
      // assumes POS will be set before POSP
      lon += lonNormalization;
      lat += latNormalization;
    }
    coords[3] = pospLon = lon;
    coords[4] = pospLat = lat;
    coords[5] = pospAlt = alt;
    pospSet = (layer.bVec == -1);
    updateMinMax();
  }

  /** Internal use only */
  private void setPosp (Object obj) {
    if (obj instanceof Table)
      setPosp((Table) obj);
    if (obj instanceof Position)
      setPosp((Position) obj);
    if (obj instanceof double[])
      setPosp((double[]) obj);
  }

  /**
   * Sets the pointing position of the point with a table.
   * @param t Pointing position table <code>{LON=lon,LAT=lat,ALT=alt}</code>
   */
  public void setPosp (Table t) {
    setPosp(t.getD("ALT", pospAlt), t.getD("LAT", pospLat), t.getD("LON", pospLon)); // default to exiting values - since 3.7.0 Bug 2965
  }

  /**
   * Sets the pointing position of the point with a Position.
   * @param p Position in {@link Position#GEODETIC} mode.
   */
  public void setPosp (Position p) {
    setPosp(p.getAlt(), p.getLat(), p.getLon());
  }

  /**
   * Sets the pointing position of the point with an array (lon,lat,alt).
   * @param a The pointing position <code>{ lon, lat, alt }</code>
   */
  public void setPosp (double[] a) {
    setPosp(a[2], a[1], a[0]);
  }

  /** Gets the pointing position. */
  public Position getPosp () {
    return Position.fromGeo(pospAlt, pospLat, pospLon);
  }


  /** Internal use only */
  void setFlags (Object flags) {
    if (flags instanceof Data)
      flags = toArray(flags, new byte[8]);
    if (flags instanceof byte[])
      setFlags((byte[]) flags);
    if (flags instanceof Table)
      setFlags((Table) flags, (byte) name.length());
  }

  /** Internal use only */
  private void setFlags (byte[] flags) {
    flagNameLen = flags[0];
    flagNamePos = flags[1];
    flagFgColor = flags[2];
    flagBgColor = flags[3];
    flagCone = flags[4];
    flagInvert = flags[5];
    flagSymCode = flags[6];
    flagSymSize = flags[7];
  }

  /** Sets the name length flag. */
  void setFlagNameLen (byte val) {
    flagNameLen = val;
  }

  /** Sets the name position flag. */
  void setFlagNamePos (byte val) {
    flagNamePos = val;
  }

  /** Sets the foreground color flag. */
  void setFlagFgColor (byte val) {
    flagFgColor = val;
  }

  /** Sets the background color flag. */
  void setFlagBgColor (byte val) {
    flagBgColor = val;
  }

  /** Sets the cone/ellipse flag. */
  void setFlagCone (byte val) {
    flagCone = val;
  }

  /** Sets the invert flag. */
  void setFlagInvert (byte val) {
    flagInvert = val;
  }

  /** Sets the symbol code flag. */
  void setFlagSymCode (byte val) {
    flagSymCode = val;
  }

  /** Sets the symbol size flag. */
  void setFlagSymSize (byte val) {
    flagSymSize = val;
  }

  /** Gets the name length flag. */
  byte getFlagNameLen () {
    return flagNameLen;
  }

  /** Gets the name position flag. */
  byte getFlagNamePos () {
    return flagNamePos;
  }

  /** Gets the foreground color flag. */
  byte getFlagFgColor () {
    return flagFgColor;
  }

  /** Gets the background color flag. */
  byte getFlagBgColor () {
    return flagBgColor;
  }

  /** Gets the cone/ellipse flag. */
  byte getFlagCone () {
    return flagCone;
  }

  /** Gets the invert flag. */
  byte getFlagInvert () {
    return flagInvert;
  }

  /** Gets the symbol code flag. */
  byte getFlagSymCode () {
    return flagSymCode;
  }

  /** Gets the symbol size flag. */
  byte getFlagSymSize () {
    return flagSymSize;
  }

  /** Sets the flags via a table. */
  private void setFlags (Table t) {
    setFlags(t, flagNameLen);
  }

  /** Internal use only. */
  private void setFlags (Table tab, byte nameLen) {
    flagNameLen = tab.getB("NAMELEN", nameLen);
    flagNamePos = tab.getB("NAMEPOS", flagNamePos);
    flagFgColor = tab.getB("COLOR", flagFgColor);
    flagBgColor = tab.getB("BGCOLOR", flagBgColor);
    flagCone    = tab.getB("CONE", flagCone);
    flagInvert  = tab.getB("INVERT", flagInvert);
    flagSymCode = tab.getB("SYMBOLID", flagSymCode);
    flagSymSize = tab.getB("SYMBOLSIZE", flagSymSize);
  }

  /** Gets the flag as a data object. */
  private Data getFlagsData () {
    byte[] array = { flagNameLen, flagNamePos, flagFgColor, flagBgColor, flagCone, flagInvert,
        flagSymCode, flagSymSize };
    return new Data(array);
  }

  /** Converts a Data object to an array. */
  private Object toArray (Object obj, Object array) {
    Data data = (Data) obj;
    data.toArray(array);
    return array;
  }

  /**
   * Sets the ellipse parameters with a table.
   * @param t ellipse parameters table <code>{SEMIMAJOR=elpsSemeiMajor,SEMIMINOR=elpsSemiMinor,TILT=elpsTilt}</code>
   * @since NeXtMidas 3.7.0 for Bug 2965
   */
  private void setElps (Table t) {
    setElps(t.getD("SEMIMAJOR", elpsSemiMajor), t.getD("SEMIMINOR", elpsSemiMinor), t.getD("TILT",elpsTilt)); // default to exiting values - since 3.7.0 Bug 2965
  }

  /**
   * Sets the ellipse parameters from an array (a,b,c).
   * @param semiMajor The semi-major axis (in meters).
   * @param semiMinor The semi-minor axis (in meters).
   * @param tilt The tilt (in deg).
   */
  public void setElps (double semiMajor, double semiMinor, double tilt) {
    elpsSemiMajor = semiMajor;
    elpsSemiMinor = semiMinor;
    elpsTilt = tilt;
    elpsSet = true;
    updateMinMax();
  }

  /**
   * Sets the ellipse parameters from an array (a,b,c).
   * @param a The ellipse parameters <code>{ semiMajor, semiMinor, tilt }</code>.
   */
  public void setElps (double[] a) {
    setElps(a[0], a[1], a[2]);
  }

  /** Internal use only. */
  private void setElps (Object obj) {
    if (obj instanceof Table) // Since 3.7.0 for bug 2965
      setElps((Table) obj);
    if (obj instanceof Data)
      obj = toArray(obj, new double[3]);
    if (obj instanceof double[])
      setElps((double[]) obj);
  }

  /**
   * Sets the cone parameters
   * @param a (saved, but not used at this time)
   * @param b (saved, but not used at this time)
   * @param angle Cone angle.
   */
  public void setCone (double a, double b, double angle) {
    coneA = a;
    coneB = b;
    coneAngle = angle;
    coneSet = true;
    updateMinMax();
  }

  /** Gets the elps as a data object. */
  private Data getElpsData () {
    double[] array = { elpsSemiMajor, elpsSemiMinor, elpsTilt };
    return new Data(array);
  }

  /**
   * Sets the cone parameters from an array (a,b,c).
   * @param a The cone parameters <code>{ a, b, angle }</code>.
   */
  public void setCone (double[] a) {
    setCone(a[0], a[1], a[2]);
  }


  /** Gets the time associated with the point */
  double getTime () {
    return time;
  }

  /** Set Time is seconds
   *  @param time double containing time in seconds (since 1950)
   *  @since NeXtMidas 3.9.3 for NXM-3515
   */
  void setTime (double time) {
//    if (time < Time.SECONDS_PER_DAY) time += Time.currentTime().getYiS();  TODO - Do we need this ...Time might take care of the csecond current day
    this.time = time;
  }

  /** Set Time is seconds
   *  @param time double containing time in seconds (since 1950)
   *  @since NeXtMidas 3.9.3 for NXM-3515
   */
  void setTime (Object obj) {
     if (obj instanceof Data)
      setTime(((Data) obj).toD());
    if (obj instanceof double[])
      setTime((double[]) obj);
  }

  Object getMetadata () {
    return metadata;
  }

  void setMetadata (Object metadata) {
    this.metadata = metadata;
  }

  /** Internal use only. */
  private void setCone (Object obj) {
    if (obj instanceof Data)
      obj = toArray(obj, new double[3]);
    if (obj instanceof double[])
      setCone((double[]) obj);
  }

  private Data getConnData () {
    return (conn) ? new Data(1) : new Data(0);
  }

  /** Gets the cone as a data object. */
  private Data getConeData () {
    double[] array = { coneA, coneB, coneAngle };
    return new Data(array);
  }

  /** Returns whether this point is hidden by another object. */
  public boolean isHidden () {
    return hidden;
  }

  /** Gets the points values based as a table. */
  Table toTable() {
    Table table = new Table();

    table.put("NAME", getName());
    table.put("NOTE", getNote());
    table.put("CONN", getConnData());
    table.put("POS", getPos());
    table.put("POSP", getPosp());
    table.put("ELPS", getElpsData());
    table.put("CONE", getConeData());
    table.put("FLAG", getFlagsData());
    table.put("TIME",Time.toString(time,Time.STD,6));
    if ((metadata !=null)) {
      table.put("METADATA", metadata); // 3.9.3
    }
    if ((layer != null) && (layer.name != null)) { // since 3.9.3
      table.put("LAYER", layer.name);
    }

    return table;
  }

  /** Sets the points values based on a table. */
  void fromTable (Table tbl) {
    setName(tbl.getS("NAME", name));
    setNote(tbl.getS("NOTE", note));

    setConn(tbl.getKey("CONN"));
    setPos(tbl.getKey("POS"));
    setPosp(tbl.getKey("POSP"));
    setFlags(tbl.getKey("FLAG"));
    setElps(tbl.getKey("ELPS"));
    setCone(tbl.getKey("CONE"));
    setTime(tbl.getKey("TIME"));
    if (!tbl.containsKey("METADATA")) {
      String pipeSeparatedMetaNames = tbl.getS("METADATAFIELDS");
      if (pipeSeparatedMetaNames != null) {
        String[] metadataFieldNames = (pipeSeparatedMetaNames==null || pipeSeparatedMetaNames.isEmpty()) ?
            null : tbl.getS("METADATAFIELDS").split("\\|");
        layer.setMetadataFieldNames(metadataFieldNames);
        metadata = new Table();
        for (String field : metadataFieldNames) {
          ((Table) metadata).put(field, tbl.getKey(field));
        }
      }
    }
    else {
      metadata = tbl.getKey("METADATA");
    }
    if (tbl.containsKey("COLOR")) {
      Object colorValue = tbl.get("COLOR");
      directColor = MidasColor.getColor(colorValue);
    }
  }

  @Override
  public String toString () {
    return toTable().toString();
  }

  /** Forms a Table describing the point, but only active attributes are included
   *  (for use in popup describing point
   *
   *  @return Table describing this point
   *  @since NeXtMidas 3.9.3
   */
  Table toDetailedTable () {
    Table table = new Table();
    if ((name!=null) && !name.isEmpty()) table.put("NAME", getName());
    if (isGeodeticData) {
      table.put("Pos Lat", posLat);
      table.put("Pos Lon", posLon);
      table.put("Pos Alt", posAlt);
      if (pospSet) {
        table.put("Posp Lat", pospLat);
        table.put("Posp Lon", pospLon);
        table.put("Posp Alt", pospAlt);
      }
    }
    else {
      table.put("Pos X", posLon);
      table.put("Pos Y", posLat);
      table.put("Pos Z", posAlt);
      if (pospSet) {
        table.put("Posp X", pospLon);
        table.put("Posp Y", pospLat);
        table.put("Posp Z", pospAlt);
      }
    }
    if (elpsSet) {
      table.put("Ellipse Major",elpsSemiMajor);
      table.put("Ellipse Minor",elpsSemiMinor);
      table.put("Ellipse Tilt",elpsTilt);
    }
    if (coneSet) {
      table.put("Cone A",coneA);
      table.put("Cone B",coneB);
      table.put("Cone Angle",coneAngle);
    }
    if (time > 0) { //ToDO decide if want this
      table.put("Time",Time.toString(time,Time.STD,6));
    }
    if ((note != null) && !note.isEmpty()) { // since 3.9.3
      table.put("Note", note);
    }
    if ((metadata !=null)) { // experimental 3.9.3
      table.put("Metadata", metadata);
    }
    if ((layer !=null) && (layer.name !=null)) { // since 3.9.3
      table.put("Layer", layer.getName());
    }
    return table;
  }

  /**
   * @return the posLat
   */
  public double getPosLat () {
    return posLat;
  }

  /**
   * @return the posLon
   */
  public double getPosLon () {
    return posLon;
  }

}
