#!/bin/bash
#################################################EXPERIMENTAL PURPOSES ONLY##################################################
#                                                                                                                           #
# A shell script that enables swapping between a few NeXtMidas swap modes:                                                  #
#   1. splitliblibg/nsplitliblibg - Handles the separation of lib-libg dependencies regarding NeXtMidas                     #
#                                   graphics-related functionality. Separation can be enabled (splitliblibg)                #
#                                   or disabled (nsplitliblibg).                                                            #
#   2.                 bcel/nbcel - As of Jdk 8+, if KeyValuePair's are being used to create lambda expressions, then       #
#                                   the bcel jar file is needed (bcel); otherwise, the bcel jar file is not needed (nbcel). #
#   3.             lucene/nlucene - If both the CLI functionality for the HELP command and the graphics functionalities for #
#                                   the HELP command are being used, then the lucene jar file is needed (lucene);           #
#                                   otherwise, if only the CLI functionality for the HELP command is being used, then       #
#                                   the lucene jar file is not needed (nlucene).                                            #
#                                                                                                                           #
# This script takes in swap mode arguments, and based on each argument, if the corresponding set of txt                     #
# files can be found within the swap parent directory of NMROOT, those files will be copied over as                         #
# into NMROOT as java files. These new java files will replace the current corresponding java files                         #
# in NMROOT with the new corresponding java files.                                                                          #
#                                                                                                                           #
# Also, if there is a jar-file dependency associated with the swap mode, then that corresponding                            #
# jar file will get moved to the appropriate directory automatically.                                                       #
#                                                                                                                           #
# Usage: <swapArg1> [swapArg2] [swapArg3]                                                                                   #
#    eg: sh $0 nsplit lucene bcel                                                                                           #
#            - Disables separation of lib-libg dependencies regarding NeXtMidas graphics-related functionality              #
#            - Adds in the lucene jar file when the graphics functionalities for the HELP command will be used              #
#            - Adds in the bcel jar file when KeyValuePair's will be used to create lambda expressions                      #
#    eg: sh $0 nbcel nlucene                                                                                                #
#            - Removes the bcel jar when KeyValuePair's will not be used to create lambda expressions                       #
#            - Removes the lucene jar file when the graphics functionalities for the HELP command will not be used          #
#    eg: sh $0 split                                                                                                        #
#            - Enables separation of lib-libg dependencies regarding NeXtMidas graphics-related functionality               #
#                                                                                                                           #
# Environment Variables:                                                                                                    #
#   NMROOT - Current NeXtMidas distribution                                                                                 #
#                                                                                                                           #
# Notes:                                                                                                                    #
# 1. At least one swap-mode argument must be provided.                                                                      #
# 2. Up to three swap-mode arguments can be provided.                                                                       #
# 3. No more than three swap-mode arguments can be provided since there are only three                                      #
#    types of swap-mode arguments available.                                                                                #
# 4. The types of swap-mode arguments can be provided in any order.                                                         #
#                                                                                                                           #
# Author: Zachary Matthews                                                                                                  #
# Since:  NeXtMidas 4.0.0                                                                                                   #
#                                                                                                                           #
#############################################################################################################################

output_help=false
valid_four_arg=false
exit_flag=false

if [ $# -lt 2 ] || [ $# -gt 3 ]; then
  if [ $# -eq 0 ] ; then
    echo "Must provide at least one swap-mode argument:"
    output_help=true
  elif [ $# -eq 1 ] ; then
    if [[ $1 == "-b" ]] || [[ $1 == "--build" ]] ; then
      echo "Must provide at least one swap-mode argument:"
      output_help=true
    fi
  elif [ $# -eq 4 ] ; then
    for cmd_arg in "$@"
    do
      if [[ ${cmd_arg} == "-b" ]] || [[ ${cmd_arg} == "--build" ]] ; then
        valid_four_arg=true
        break
      fi
    done
    if ! ${valid_four_arg} ; then
      echo "No more than three swap-mode arguments are allowed:"
      output_help=true
    fi
  else
    echo "No more than three swap-mode arguments are allowed:"
    output_help=true
  fi
  
  if ${output_help} ; then
    echo "Usage: sh $0 <swapArg1> [swapArg2] [swapArg3]"
    echo "   eg: sh $0 nsplit lucene bcel"
    echo "   eg: sh $0 nbcel nlucene"
    echo "   eg: sh $0 split"
    echo ""
    echo "optional flags:"
    echo "-b --build     rebuild NeXtMidas after swapping java files and jar files"
    echo ""
    exit_flag=true
  fi
fi

# Must have NMROOT defined
if [ -z "${NMROOT}" ] ; then
  echo "NMROOT is not set, cannot swap files"
  echo ""
  exit_flag=true
fi

if ${exit_flag} ; then
  exit 1
fi

orig_dir=${PWD}                # original directory
swap_parent_dir=${NMROOT}/swap # swap parent directory

rebuild_nxm=false

nxm_repo=`basename ${NMROOT}`
echo "Swapping for NeXtMidas repo ${nxm_repo}"
for arg in "$@"
do
  if [[ ${arg} == "-b" ]] || [[ ${arg} == "--build" ]] ; then
    rebuild_nxm=true
    continue
  fi

  cd ${swap_parent_dir} # get back into the swap parent directory 
  swap_mode_dir=${arg}
  # **********************************************************************
  # Determine swap subdirectory corresponding to the
  # current input argument (swapArg) from the user
  # **********************************************************************
  swap_mode_dir_array=(`find . -mindepth 2 -maxdepth 2 -type d -iname "*${swap_mode_dir}*"`) # assuming directory names are always 2 levels below swap parent directory
  if [ ${#swap_mode_dir_array[@]} -eq 0 ] ; then   # if no entries are found, then assume invalid argument
    echo "Invalid argument: ${swap_mode_dir}, no such swap mode. Cannot swap files for ${swap_mode_dir}."
    echo ""
    continue
  elif [ ${#swap_mode_dir_array[@]} -eq 1 ] ; then # if only one entry is found, then assume it begins with 'n'
    swap_mode_dir=${swap_mode_dir_array[0]}
    need_jar=false
  else                                             # if two entries are found, then assume one entry is one character shorter ('n') than the other
    if [ ${#swap_mode_dir_array[0]} -lt ${#swap_mode_dir_array[1]} ] ; then
      swap_mode_dir=${swap_mode_dir_array[0]}
    else
      swap_mode_dir=${swap_mode_dir_array[1]}
    fi
    need_jar=true
  fi

  # **********************************************************************
  # Copy all files within the swap subdirectory to the temp directory,
  # change them to java files, and move the java files into NMROOT
  # to replace the corresponding files currently in NMROOT
  # **********************************************************************
  mode_name=`basename ${swap_mode_dir}`
  echo "Swapping files for ${mode_name} mode..."
  cd ${swap_mode_dir}
  files_to_copy=`find . -type f -name "*.txt"` # assuming text files here in mode directory
  for file in ${files_to_copy}
  do
    opt_tree=`cut -d '/' -f2 <<< ${file}`           # get top-level option tree
    text_file=${file:2}                             # name of the text file without "./"
    java_file=${text_file::-4}.java                 # name of the java-file version of the text file
    cp --parents ${text_file} /tmp                  # preserve original parent directory structure (sys/...) 
    mv /tmp/${text_file} ${NMROOT}/nxm/${java_file} # convert files to java files as they're being moved over; for 'mv', files are overwritten in NMROOT by default
  done
  rm -rf /tmp/${opt_tree} # remove /tmp/<option_tree>
  echo "...${mode_name} swapping files complete"
  echo ""

  # **********************************************************************
  # Move jar file into swap directory or vice versa (if applicable)
  # **********************************************************************
  echo "Checking for jars to be moved for ${mode_name} mode..."
  nxm_opt_jars_dir=${NMROOT}/nxm/${opt_tree}/jars
  nmroot_jars_dir=${NMROOT}/jars
  if ${need_jar} ; then                                             # move jar back into nxm/opt_tree/jars
    cd ..
    jar_file=`find . -maxdepth 1 -type f -iname "*${mode_name}*.jar"`   # assumed to be of length 1
    if [ -z ${jar_file} ] ; then                                      # jar is assumed to already be back in nxm/opt_tree/jars and $NMROOT/jars
                                                                      # or to not have existed to begin with
      echo "...no jars to be moved for ${mode_name} mode"
    else                                                              # jar file must be moved into nxm/opt_tree/jars
      if [ ! -d ${nxm_opt_jars_dir} ] ; then
        mkdir -p ${nxm_opt_jars_dir}                                    # make nxm/opt/jars directory
      fi
      mv ${jar_file} ${nxm_opt_jars_dir}                              # move jar file into nxm/opt/jars
      echo "...added ${jar_file:2} back into ${nxm_opt_jars_dir}"
    fi
  else                                                              # remove jar from nxm/opt_tree/jars and nmroot/jars
    jars_dirs_array=(${nxm_opt_jars_dir} ${nmroot_jars_dir})          # contains both jar directories
    for jar_dir in "${jars_dirs_array[@]}"                            # loop through each jar directory
    do
      if [ -d ${jar_dir} ] ; then                                  # jar directory has jars
        cd ${jar_dir}
        jar_file=`find . -maxdepth 1 -type f -iname "*${mode_name:1}*.jar"` # assumed to be of length 1
        if [ -z ${jar_file} ] ; then                                        # jar is assumed to have already been removed from jar directory
          echo "...no jars to be moved from ${jar_dir} for ${mode_name} mode"
        else                                                                # jar file must be removed from jar directory
          mv ${jar_file} ${swap_parent_dir}/${swap_mode_dir}/..
          echo "...removed ${jar_file:2} from ${jar_dir}"
        fi
      else                                                              # no jars directory means there are no jars in the corresponding option tree to begin with
        echo "...no jars to be moved from ${jar_dir} for ${mode_name} mode"
      fi
    done
  fi
  echo ""
done

cd ${orig_dir} # jump back into the original directory

# **********************************************************************
# Clean and rebuild NeXtMidas repo
# **********************************************************************
if ${rebuild_nxm} ; then
  pid=$$
  transcript="/tmp/transcript_${pid}.txt"
  echo "Cleaning and rebuilding NeXtMidas repo ${nxm_repo}..."
  source ${NMROOT}/os/unix/nm.sh make clean all > /dev/null
  echo "...cleaning complete..."
  source ${NMROOT}/os/unix/nm.sh make all >& ${transcript}
  transcript_build_errors=(`egrep "BUILD FAILED|[0-9]+ error|Unable to run NeXtMidas command" ${transcript}`)

  # Check whether NeXtMidas build passed or failed 
  if [ ${#transcript_build_errors[@]} -gt 0 ] ; then
    cat ${transcript}
    echo ""
    echo "...rebuilding failed (see output above)"
  else
    echo "...rebuilding complete"
  fi

  echo ""
  rm ${transcript} # remove temporary transcript file
fi
